@@ -1,7 +1,7 @@
 /*
 	ccid_usb.c: USB access routines using the libusb library
 	Copyright (C) 2003-2010	Ludovic Rousseau
-	Copyright (C) 2009-2017	Advanced Card Systems Ltd.
+	Copyright (C) 2009-2018	Advanced Card Systems Ltd.
 
 	This library is free software; you can redistribute it and/or
 	modify it under the terms of the GNU Lesser General Public
@@ -693,6 +693,7 @@ again:
 				usb_interface = get_ccid_usb_interface(config_desc, &num);
 				if (usb_interface == NULL)
 				{
+					libusb_free_config_descriptor(config_desc);
 					(void)libusb_close(dev_handle);
 					if (0 == num)
 						DEBUG_CRITICAL3("Can't find a CCID interface on %d/%d",
@@ -715,6 +716,7 @@ again:
 						(readerID != ACS_ACR1281_1S_DUAL_READER) &&
 						(readerID != ACS_ACR1281_2S_CL_READER))
 					{
+						libusb_free_config_descriptor(config_desc);
 						(void)libusb_close(dev_handle);
 						DEBUG_CRITICAL3("Unable to find the device descriptor for %d/%d",
 							bus_number, device_address);
@@ -726,6 +728,7 @@ again:
 				interface = usb_interface->altsetting->bInterfaceNumber;
 				if (interface_number >= 0 && interface != interface_number)
 				{
+					libusb_free_config_descriptor(config_desc);
 					/* an interface was specified and it is not the
 					 * current one */
 					DEBUG_INFO3("Found interface %d but expecting %d",
@@ -742,6 +745,7 @@ again:
 				r = libusb_claim_interface(dev_handle, interface);
 				if (r < 0)
 				{
+					libusb_free_config_descriptor(config_desc);
 					(void)libusb_close(dev_handle);
 					DEBUG_CRITICAL4("Can't claim interface %d/%d: %s",
 						bus_number, device_address, libusb_error_name(r));
@@ -758,6 +762,7 @@ again:
 				/* check for firmware bugs */
 				if (ccid_check_firmware(&desc))
 				{
+					libusb_free_config_descriptor(config_desc);
 					(void)libusb_close(dev_handle);
 					return_value = STATUS_UNSUCCESSFUL;
 					goto end2;
@@ -976,6 +981,8 @@ again:
 #endif
 					usbDevice[reader_index].multislot_extension = NULL;
 
+				libusb_free_config_descriptor(config_desc);
+
 				// Get number of slots
 				numSlots = usbDevice[reader_index].ccid.bMaxSlotIndex + 1;
 
@@ -1096,6 +1103,10 @@ end:
 			goto again_libusb;
 		}
 #endif
+
+		/* free bundle list */
+		bundleRelease(&plist);
+
 		/* failed */
 		close_libusb_if_needed();
 
@@ -1832,9 +1843,9 @@ int InterruptRead(int reader_index, int timeout /* in ms */)
 
 		default:
 			/* if libusb_interrupt_transfer() times out we get EILSEQ or EAGAIN */
-			DEBUG_COMM4("InterruptRead (%d/%d): %s",
+			DEBUG_COMM4("InterruptRead (%d/%d): %d",
 				usbDevice[reader_index].bus_number,
-				usbDevice[reader_index].device_address, libusb_error_name(ret));
+				usbDevice[reader_index].device_address, ret);
 			return_value = IFD_COMMUNICATION_ERROR;
 	}
 
@@ -1866,11 +1877,41 @@ void InterruptStop(int reader_index)
 
 		ret = libusb_cancel_transfer(transfer);
 		if (ret < 0)
-			DEBUG_CRITICAL2("libusb_cancel_transfer failed: %d", ret);
+			DEBUG_CRITICAL2("libusb_cancel_transfer failed: %s",
+				libusb_error_name(ret));
 	}
 } /* InterruptStop */
 
 
+/*****************************************************************************
+ *
+ *					TriggerSlotChange
+ *
+ ****************************************************************************/
+void TriggerSlotChange(int reader_index)
+{
+	struct usbDevice_MultiSlot_Extension *msExt =
+		usbDevice[reader_index].multislot_extension;
+
+	if (msExt != NULL)
+	{
+		int interrupt_byte =
+			(usbDevice[reader_index].ccid.bCurrentSlotIndex / 4) + 1;
+		int interrupt_value =
+			0x02 << (2 * (usbDevice[reader_index].ccid.bCurrentSlotIndex % 4));
+
+		pthread_mutex_lock(&msExt->mutex);
+
+		msExt->status = LIBUSB_TRANSFER_COMPLETED;
+		memset(msExt->buffer, 0, sizeof(msExt->buffer));
+		msExt->buffer[interrupt_byte] = interrupt_value;
+
+		pthread_cond_broadcast(&msExt->condition);
+		pthread_mutex_unlock(&msExt->mutex);
+	}
+}
+
+
 /*****************************************************************************
  *
  *					Multi_PollingProc
